%% Maze Environment

classdef maze
    properties
        snum % Number of states
        anum % Number of actions
        fnum % Number of flags
        mapDim % Map size (rows, columns)
        obstacles % Fixed locations of the obstacles
        flagsLoc % Fixed locations of the obstacles
        collectedFlags % 1x3, if a flag has been collected, the corresponding element is 1. 0 otherwise
        flagsList % Possible combinations of the flag collection
        slipProb % Slip probability
        agentPos % Agent Position
        goalPos % Goal Position
        idx2cell % Potential cells agent can be on
        map % Map visualization
        actions % Action indices
        slipActions % Corresponding slip action indices
        actionMov % Action Movement
 
    end
    methods
        % Constructor
        function obj = maze()
            obj.snum = 112;
            obj.anum = 4;
            obj.fnum = 3;
            obj.slipProb = 0.1;
            obj.mapDim = [4,5];
            
            % 1:Up, 2: Down, 3: Left, 4: Right
            obj.actions = 1:obj.anum; 
            
            % When slip, move the the next clockwise action
            % 1->4; 2->3, 3->1, 4->2
            obj.slipActions = [4,3,1,2];
            
            % Action movements
            obj.actionMov = [[-1,0];[1,0];[0,-1];[0,1]];
            
            obj.obstacles = [[1,2];[1,4];[3,1];[3,5];[4,3];[4,5]];
            obj.goalPos = [1,5];
            obj.flagsLoc = [[1,3];[4,1];[4,4]];
            obj.flagsList = [[0,0,0];[1,0,0];[0,1,0];[0,0,1];[1,1,0];[1,0,1];[0,1,1];[1,1,1]];
            obj.idx2cell = [[1,1];[2,1];[4,1];[2,2];[3,2];[4,2];[1,3];...
                            [2,3];[3,3];[2,4];[3,4];[4,4];[1,5];[2,5]];
            % Initialize map
            obj.map = zeros(obj.mapDim);
            obj.map(1,5) = 2;
            for i=1:length(obj.obstacles)
                obj.map(obj.obstacles(i,1),obj.obstacles(i,2)) = 4;
            end
            
            % Reset environment
            obj = obj.reset();
        end
        
        % Reset environment to starting state
        function obj = reset(obj)
            obj.agentPos = [1,1];
            obj.collectedFlags = [0,0,0];
        end
        
        % Set environment to given state
        function obj = set_state(obj, state)
            % Implement your code here.
        end
        
        % Get current state of environment
        function state = get_state(obj)
            % Implement your code here.
        end
        
        % Helper function to return valid states
        % Some states are never reached (invalid)
        % i.e. when agent is on a cell and flag is present on cell
        function valid = get_valid_states(obj)
            valid = ones(obj.snum,1);
            for i = 1:obj.snum
                tempAgentPos = obj.idx2cell(floor((i-1)/8)+1,:);
                tempCollectedFlags = obj.idx2flag(mod((i-1),8)+1);
                remainingFlags = obj.flagsLoc(logical(1-tempCollectedFlags),:);
                if any(prod(remainingFlags == tempAgentPos,2)) || prod(obj.goalPos == tempAgentPos,2)
                    valid(i) = 0;
                end
            end
        end
            
        % Set the slip probability
        function obj = setSlipProb(obj, prob)
            obj.slipProb = prob;
        end
        
        % Move according to actionId
        function [obj, reward, terminal] = move(obj, actionId)
            if rand() < obj.slipProb
                exec_ac = obj.slipActions(actionId);
            else
                exec_ac = actionId;
            end
            nextPos = obj.agentPos + obj.actionMov(exec_ac,:);
            reward = 0;
            terminal = 0;
            if ~checkCollisions(obj, nextPos)
                if prod(nextPos == obj.goalPos,2) == 1
                    terminal = 1;
                    reward = sum(obj.collectedFlags);
                    obj = reset(obj);
                else
                    terminal = 0;
                    reward = 0;
                    obj.collectedFlags(find(prod(obj.flagsLoc == nextPos,2)==1))=1;
                    obj.agentPos = nextPos;
                end
            end
        end
        
        % Check collisions
        function collision = checkCollisions(obj, pos)
            obsCollision = prod(pos==obj.obstacles,2);
            if sum(obsCollision ~= 0)
                collision = true;
            elseif any(pos < 1) ||...
                    pos(1) > obj.mapDim(1) ||...
                    pos(2) > obj.mapDim(2)
                collision = true;
            else
                collision = false;
            end
        end
        
        % Convert flag combination from idx
        function flag_vec = idx2flag(obj, n)
            % n = 1,...,8
            flag_vec = obj.flagsList(n,:);
        end
        
        % Get idx from current flag combination
        function idx = flag2idx(obj)
            if sum(obj.collectedFlags) < 2
                idx = obj.collectedFlags*[1,2,3]'+1;
            else
                idx = obj.collectedFlags*[1,2,3]'+2;
            end
        end
        
        % Visualization
        function visualize(obj)
           img = obj.map;
           img(img==1)=0;
           img(img==2)=0;
           img(img==3)=0;
           colormap(flipud(gray));
           imagesc(img);
           
           currFlagsLoc = obj.flagsLoc(find(obj.collectedFlags == 0),:);
           fstring = repmat(['F'], size(currFlagsLoc,1),1);
           agentString = text(obj.agentPos(:,2),obj.agentPos(:,1),...
                ['A'], 'HorizontalAlignment', 'center');
           goalString = text(obj.goalPos(:,2),obj.goalPos(:,1),...
                ['G'], 'HorizontalAlignment', 'center');
           flagString = text(currFlagsLoc(:,2),currFlagsLoc(:,1),...
                fstring, 'HorizontalAlignment', 'center');
           set(agentString, 'Color', 'b', 'FontSize', 20);
           set(goalString, 'Color', 'g', 'FontSize', 20);
           set(flagString, 'Color', 'r', 'FontSize', 20);
           
           k = waitforbuttonpress;
        end
    end
end